from pipelines.io import make_directory
from Queue import Queue

import os
import subprocess
import threading

class JobManager(object):
    def add_job(self, cmd, cmd_args, **kwargs):
        '''
        Add a job to a queue to be run.
        '''
        pass
    
    def run_job(self, cmd, cmd_args, **kwargs):
        '''
        Run a job and wait for it to complete.
        '''
        pass    
    
    def wait(self, job):
        '''
        Wait for all jobs in the queue to finish.
        '''
        pass
    
    def close(self):
        '''
        Close the job manager.
        '''
        pass
    
class LocalJobManager(JobManager):
    def __init__(self):
        self.cmd_strs = []
    
    def add_job(self, cmd, cmd_args, **kwargs):
        cmd_str = ' '.join([cmd, ] + [str(x) for x in cmd_args])
        
        self.cmd_strs.append(cmd_str)
    
    def run_job(self, cmd, cmd_args, **kwargs):
        cmd_str = ' '.join([cmd, ] + [str(x) for x in cmd_args])
        
        self._run_cmd(cmd_str)
    
    def wait(self):
        for cmd_str in self.cmd_strs:
            self._run_cmd(cmd_str)
            
    def _run_cmd(self, cmd_str):
        '''
        Throw exception if run command fails
        '''
        process = subprocess.Popen(cmd_str, stdin=subprocess.PIPE, shell=True)
       
        process.stdin.close()
        
        sts = os.waitpid(process.pid, 0)    
        
        if sts[1] != 0:
            raise Exception('Failed to run {0}\n'.format(cmd_str))

class ClusterJobManager(JobManager):
    def __init__(self, log_dir=None, max_tries=1):
        import drmaa
        
        self._drmaa = drmaa
        
        self.log_dir = log_dir
        
        self.max_tries = max_tries
        
        if self.log_dir is not None:
            make_directory(self.log_dir)
        
        self.session = self._drmaa.Session()

        self.session.initialize()

        self.jobs = Queue()
        
        self._lock = threading.Lock()
    
    def add_job(self, cmd, cmd_args, mem=4, max_mem=10, num_cpus=1):
        job = ClusterJob(cmd, cmd_args, mem, max_mem, num_cpus)
        
        self.jobs.put(job)
        
    def run_job(self, cmd, cmd_args, mem=4, max_mem=10, num_cpus=1):
        job = ClusterJob(cmd, cmd_args, mem, max_mem, num_cpus)
        
        num_tries = 0
        
        while True:
            num_tries += 1
            
            job_id = self._run_job(job)
            
            exit_status = self._check_exit_status(job_id)
            
            if exit_status == 0:
                break
            
            elif num_tries > self.max_tries:
                raise Exception('Job {0} failed with exit status {1}.'.format(job_id, exit_status))
                
    def wait(self):
        jobs = self._get_jobs()
        
        num_tries = 0
        
        while True:
            num_tries += 1
            
            job_map = self._run_multiple_jobs(jobs)
            
            exit_statuses = self._check_multiple_exit_statuses(job_map.keys())
            
            failed_jobs = []
            
            for job_id in job_map:
                if exit_statuses[job_id] != 0:
                    failed_jobs.append(job_map[job_id])
               
            if len(failed_jobs) == 0:
                break
            
            elif num_tries < self.max_tries:
                jobs = failed_jobs
            
            else:
                error_strings = []
                
                for job_id in job_map:
                    if exit_statuses[job_id] != 0:
                        error_strings.append('Job {0} failed with exit status {1}.'.format(job_id, exit_statuses[job_id]))
                
                raise Exception('\n'.join(error_strings))

    def close(self):
        self.session.control(self._drmaa.Session.JOB_IDS_SESSION_ALL, self._drmaa.JobControlAction.TERMINATE)
        
        self.session.exit()

    def _check_exit_status(self, job_id):
        return_value = self.session.wait(job_id, self._drmaa.Session.TIMEOUT_WAIT_FOREVER)
        
        return return_value.exitStatus

    def _check_multiple_exit_statuses(self, job_ids):
        '''
        Check the exit statuses of a list of job IDs.
        '''
        exit_statuses = {}

        for job_id in job_ids:
            exit_statuses[job_id] = self._check_exit_status(job_id)
        
        return exit_statuses
    
    def _get_jobs(self):
        '''
        Get list of pending jobs in a thread safe manner.
        '''
        self._lock.acquire()
        
        jobs = []
        
        while not self.jobs.empty():
            jobs.append(self.jobs.get())
        
        self._lock.release()
        
        return jobs
        
    def _run_job(self, job):      
        job_template = job.get_job_template(self.log_dir, self.session)
        
        job_id = self.session.runJob(job_template)
        
        self.session.deleteJobTemplate(job_template)
        
        return job_id
    
    def _run_multiple_jobs(self, jobs):
        '''
        Run a list of jobs returning a mapping of job IDs -> jobs.
        '''
        job_map = {}
        
        for job in jobs:
            job_id = self._run_job(job)
            
            job_map[job_id] = job

        self.session.synchronize(job_map.keys(), self._drmaa.Session.TIMEOUT_WAIT_FOREVER, False)
        
        return job_map
              
class ClusterJob(object):
    def __init__(self, cmd, cmd_args, mem, max_mem, num_cpus):
        self.cmd = cmd
        
        self.cmd_args = cmd_args
        
        self.mem = mem
        
        self.max_mem = max_mem
        
        self.num_cpus = num_cpus

    def get_job_template(self, log_dir, session):
        job_template = session.createJobTemplate()
        
        job_template.remoteCommand = self.cmd
        
        job_template.args = [str(x) for x in self.cmd_args]
        
        job_template.workingDirectory = os.getcwd()
        
        if log_dir is not None:
            job_template.errorPath = ':' + log_dir
            
            job_template.outputPath = ':' + log_dir
        
        job_template.nativeSpecification = '-l mem_free={mem}G,mem_token={mem}G,h_vmem={max_mem}G -V -w n -pe ncpus {num_cpus}'.format(**self.__dict__)
        
        return job_template        

class RemoteJobManager(object):
    def __init__(self, host_name, ssh_private_key, ssh_user_name):
        import spur
        
        self._spur = spur
        
        self.host_name = host_name
        
        self.ssh_private_key = ssh_private_key
        
        self.ssh_user_name = ssh_user_name
        
        self.cmds = []
        
    def add_job(self, cmd, cmd_args, **kwargs):
        self.cmds.append((cmd, cmd_args))
    
    def run_job(self, cmd, cmd_args, **kwargs):
        self._run_cmd(cmd, cmd_args)
    
    def wait(self):
        for cmd, cmd_args in self.cmds:
            self._run_cmd(cmd, cmd_args)        
        
    def _run_cmd(self, cmd, cmd_args):
        shell = self._get_shell()
        
        results = shell.run([cmd, ] + cmd_args)
        
        if results.return_code != 0:
            raise results.to_error()

    def _get_shell(self):
        shell = self._spur.SshShell(hostname=self.host_name,
                                    private_key_file=self.ssh_private_key,
                                    username=self.ssh_user_name)
        
        return shell
        
